# SPDX-FileCopyrightText: 2022-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
#   PROGRAM/MODULE: saturnin
#   FILE:           saturnin/_scripts/repl.py
#   DESCRIPTION:    REPL for Typer application
#   CREATED:        05.08.2022
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Initial code is based on https://github.com/click-contrib/click-repl
#
# Contributor(s): Pavel Císař (initial code)
#                 ______________________________________
# pylint: disable=W0212

"""REPL for Typer application
"""

from __future__ import annotations
from typing import Dict, Any, List, TextIO, Callable, Optional
from pathlib import Path
from operator import attrgetter
import shlex
import sys
from io import StringIO
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import prompt
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
import click
from click.exceptions import Exit as ClickExit, Abort as ClickAbort
from rich.console import Console
from saturnin.base import RestartError, RESTART, directory_scheme
from saturnin.lib.console import console as cm, FORCE_TERMINAL

EchoCallback = Callable[[str], None]

#: Prompt-toolkit key bindings
kb = KeyBindings()

@kb.add('c-space')
def _(event):
    " Initialize autocompletion, or select the next completion. "
    buff = event.app.current_buffer
    if buff.complete_state:
        buff.complete_next()
    else:
        buff.start_completion(select_first=False)

class CustomClickCompleter(Completer):
    """Custom prompt-toolkit completer.

    It provides command completion for Typer/Click commands and parameters, including
    option/parameter values.

    Arguments:
      cli: Root Typer command group
    """
    def __init__(self, cli):
        self.cli = cli

    def get_completions(self, document, complete_event=None):
        """Yields completion choices.
        """
        # Code analogous to click._bashcomplete.do_complete
        try:
            txt = document.text_before_cursor
            in_help: bool = txt.startswith('?')
            if in_help:
                txt = txt[1:]
            args = shlex.split(txt)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return

        cursor_within_command = (document.text_before_cursor.rstrip()
                                 == document.text_before_cursor)

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ""
        ctx = click.shell_completion._resolve_context(self.cli, {}, "", args)
        if ctx is None:
            return

        last_arg = args[-1] if args else ''

        choices = []
        stop: bool = in_help
        if isinstance(ctx.command, click.MultiCommand):
            # Completion is list of commands at given context level
            if not args:
                choices.append(Completion('quit', -len(incomplete),
                                          display_meta="Quit Saturnin console"))
            for name in ctx.command.list_commands(ctx):
                command = ctx.command.get_command(ctx, name)
                if not command.hidden:
                    choices.append(Completion(str(name),-len(incomplete),
                                              display_meta=command.get_short_help_str()))
            stop = stop or choices
        if not stop:
            # First check whether we're entering value for option.
            for param in ctx.command.params:
                if (isinstance(param, click.Option)
                    and not param.is_flag
                    and (last_arg in param.opts or last_arg in param.secondary_opts)):
                    # Completion are possible values for last option, if applicable
                    if isinstance(param.type, click.Choice):
                        for choice in param.type.choices:
                            choices.append(Completion(str(choice), -len(incomplete)))
                    else:
                        choices.extend(Completion(str(item.value), -len(incomplete),
                                                  display_meta=item.help)
                                       for item in param.shell_complete(args, incomplete))
                    stop = True # Do not continue even if we don't have choices!
            stop = stop or choices
        if not stop:
            # We're looking for possible argument values or option
            # First we build list of already processed options and arguments...
            not_processed_params = []
            for param in ctx.command.params:
                if isinstance(param, click.Option):
                    if ctx.params[param.name] == param.default:
                        not_processed_params.append(param)
            if not incomplete.startswith('-'):
                for param in ctx.command.params:
                    if isinstance(param, click.Argument):
                        if (param.nargs == 1) and (ctx.params[param.name] == param.default):
                            not_processed_params.append(param)
                            break
                        elif param.nargs == -1:
                            not_processed_params.append(param)
                            break
            #
            for param in not_processed_params:
                if isinstance(param, click.Option):
                    # Completion is list of options
                    for options in (param.opts, param.secondary_opts):
                        for opt in options:
                            choices.append(Completion(str(opt), -len(incomplete),
                                                      display_meta=param.help))
                elif isinstance(param, click.Argument):
                    # Completion are values for argument, if applicable
                    if isinstance(param.type, click.Choice):
                        for choice in param.type.choices:
                            choices.append(Completion(str(choice), -len(incomplete),
                                                      display_meta=param.help))
                    else:
                        choices.extend(Completion(str(item.value), -len(incomplete),
                                                  display_meta=item.help if item.help
                                                  else param.help)
                                       for item in param.shell_complete(args, incomplete))
        stop = stop or choices

        choices.sort(key=attrgetter('text'))
        for item in choices:
            if item.text.startswith(incomplete):
                yield item

class IOManager: # pylint: disable=R0902
    """REPL I/O manager.

    Handles command prompt, stdin/stdout redirection etc.

    Arguments:
      context: Current Click context
      echo:    Callback called with command line before it's executed.
      console: Costom Rich console for output. If not provided, Saturnin standard console
               is used.
    """
    def __init__(self, context, *, echo: Optional[EchoCallback]=None, console: Console=None):
        self.console: Console = cm.std_console if console is None else console
        self.html_output: bool = False
        self.output_file: TextIO = None
        self.output_filename: Path = None
        self.echo: Optional[EchoCallback] = echo
        self.run_commands: List[str] = []
        self.isatty: bool = sys.stdin.isatty()
        self.saved_stdin = sys.stdin
        self.saved_stdout = sys.stdout
        self.pipe_in = StringIO()
        self.pipe_out = StringIO()
        self.prompt_kwargs: Dict[str, Any] = {}
        group_ctx = context.parent or context
        defaults = {
            'history': FileHistory(str(directory_scheme.history_file)),
            'completer': CustomClickCompleter(group_ctx.command),
            'message': '> ',
            'key_bindings': kb,
            'auto_suggest': AutoSuggestFromHistory()
        }
        for key, default_value in defaults.items():
            if key not in self.prompt_kwargs:
                self.prompt_kwargs[key] = default_value
        #
        self.cmd_queue = []
    def __enter__(self) -> IOManager:
        return self
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        sys.stdin = self.saved_stdin
        sys.stdout = self.saved_stdout
        self.restore_console()
    def _is_internal_cmd(self, cmd: str) -> bool:
        cmd = cmd.rstrip().split(' ')[0]
        if cmd.lower() in ['help', 'quit']:
            return True
        if cmd.startswith('?'):
            return True
        return False
    def _get_next_cmd(self) -> str:
        command = self.cmd_queue.pop(0)
        self.pipe_in = self.pipe_out
        self.pipe_in.seek(0)
        sys.stdin = self.pipe_in
        if self.cmd_queue:
            self.pipe_out = StringIO()
            sys.stdout = self.pipe_out
        else:
            sys.stdout = self.saved_stdout
        return command
    def _get_command(self) -> str:
        "Returns next command fetched from queue, stdin or console prompt."
        sys.stdin = self.saved_stdin
        sys.stdout = self.saved_stdout
        self.pipe_out = StringIO()
        if self.run_commands:
            command = self.run_commands.pop(0)
        elif not self.isatty:
            command = sys.stdin.readline()
        else:
            command = prompt(**self.prompt_kwargs)
        return command
    def get_command(self) -> str:
        "Returns next command."
        if self.cmd_queue:
            return self._get_next_cmd()
        command = self._get_command()
        if self.echo and command.strip():
            self.echo(command)
        sys.stdin = self.saved_stdin
        sys.stdout = self.saved_stdout
        return command
    def reset_queue(self) -> None:
        "Clear command queue"
        i = len(self.cmd_queue)
        self.cmd_queue.clear()
        sys.stdin = self.saved_stdin
        sys.stdout = self.saved_stdout
        if i > 0:
            self.console.print(f'Remaining {i} command(s) not executed')
    def redirect_console(self, filename: Path) -> None:
        """Redirects console output to file.

        Arguments:
          filename: File for console output.
        """
        if self.output_file is not None:
            self.output_file.close()
        self.output_file = filename.open(mode='w', encoding='utf8')
        self.output_filename = filename
        self.html_output = filename.suffix.startswith('.htm')
        self.console = Console(file=self.output_file, width=5000, force_terminal=FORCE_TERMINAL,
                               emoji=False, record=self.html_output)
    def restore_console(self) -> None:
        "Closes the output file and restores output to console."
        if self.output_file is not None:
            self.output_file.close()
            self.output_file = None
            if self.html_output:
                self.console.save_html(self.output_filename)
        self.console = cm.std_console

def repl(context, ioman: IOManager) -> bool: # pylint: disable=R0912
    """
    Start an interactive shell. All subcommands are available in it.

    Arguments:
      context: Current Click context.
      ioman:   IOManager instance.

    Returns:
       True if REPL should be restarted, otherwise returns False.

    If stdin is not a TTY, no prompt will be printed, but commands are read
    from stdin.
    """
    group_ctx = context.parent or context
    group_ctx.info_name = ''
    group = group_ctx.command
    group.params.clear()
    while True:
        try:
            command = ioman.get_command()
        except KeyboardInterrupt:
            continue
        except EOFError:
            break
        if not command:
            if ioman.isatty:
                continue
            break
        # Internal commands
        if ioman._is_internal_cmd(command):
            cmd = command.rstrip()
            if cmd.lower() == 'help':
                command = '--help'
            elif cmd.startswith('?'):
                command = cmd[1:] + ' --help'
            elif cmd.lower() == 'quit':
                break
        # Special commands
        for cmd in ('pip ', 'install package ', 'uninstall package '):
            if command.startswith(cmd):
                command = command[:len(cmd)] + ' -- ' + command[len(cmd):]
                break
        try:
            args = shlex.split(command)
            ctx = click.shell_completion._resolve_context(group, {}, "", args)
        except ValueError as exc:
            ioman.console.print(f"{type(exc).__name__}: {exc}")
            continue
        try:
            with group.make_context(None, args, parent=group_ctx) as ctx:
                result = group.invoke(ctx)
                #ctx.exit()
                if result is RESTART:
                    raise RestartError
        except click.ClickException as exc:
            exc.show()
            ioman.reset_queue()
        except ClickExit as exc:
            pass
            #sys.exit(exc.exit_code)
        except ClickAbort as exc:
            cm.print_error("Aborted!")
            sys.exit(1)
        except SystemExit:
            pass
        except RestartError:
            return True
        except Exception as exc: # pylint: disable=W0703
            cm.print_error(f"{exc.__class__.__name__}:{exc!s}")
    return False



# SPDX-FileCopyrightText: 2021-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/client.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Service clients
# CREATED:        15.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin base module for implementation of Firebird Butler Service clients


"""

from __future__ import annotations
import os
import platform
from struct import pack
from uuid import UUID
from saturnin.base import (Error, ZMQAddress, Channel, TIMEOUT, INVALID, AgentDescriptor,
     PeerDescriptor)
from saturnin.protocol.fbsp import (FBSPClient, FBSPSession, FBSPMessage,
     WelcomeMessage, ErrorMessage, MsgType)

class Token():
    """FBSP message token generator.
    """
    def __init__(self, value: int=0):
        self._value: int = value
    def next(self) -> bytes:
        """Returns next message token.
        """
        result = pack('!Q', self._value)
        self._value += 1
        return result

class ServiceClient:
    """Base class for Firebird Butler Service clients.
    """
    def __init__(self, timeout: int=1000):
        """
        Arguments:
            timeout: The timeout (in milliseconds) to wait for message.
        """
        self.token: Token = Token()
        self.timeout: int = timeout
        self.channel: Channel = None
        self.session: FBSPSession = None
        self.protocol: FBSPClient = None
    def open(self, channel: Channel, address: ZMQAddress, agent: AgentDescriptor,
             peer_uid: UUID) -> None:
        """Open connection to Firebird Butler service.

        Arguments:
            channel: Channel used for communication with service.
            address: Service endpoint address.
            agent: Client agent identification.
            peer_uid: Client peer ID.
        """
        assert isinstance(channel.protocol, FBSPClient)
        self.channel = channel
        self.session = channel.connect(address)
        self.protocol = channel.protocol
        self.protocol.send_hello(channel, self.session, agent,
                                 PeerDescriptor(peer_uid, os.getpid(), platform.node()),
                                 self.token.next())
        msg = self.channel.receive(self.timeout)
        if isinstance(msg, ErrorMessage):
            raise self.protocol.exception_for(msg)
        if msg is TIMEOUT:
            raise TimeoutError()
        if msg is INVALID:
            raise Error("Invalid response from service")
        if not isinstance(msg, WelcomeMessage):
            raise Error(f"Unexpected {msg.msg_type.name} message from service")
    def send(self, msg: FBSPMessage) -> None:
        """Send message to the service.

        Arguments:
            msg: Message to be sent.
        """
        self.channel.send(msg, self.session)
    def receive(self) -> FBSPMessage:
        """Receive one message from service.

        Raises:
            TimeoutError: When timeout expires.
            Error: When `Channel.receive()` returns `.INVALID` sentinel, or service closes
                   connection with CLOSE message.
        """
        msg = self.channel.receive(self.timeout)
        if isinstance(msg, ErrorMessage):
            raise self.protocol.exception_for(msg)
        if msg is TIMEOUT:
            raise TimeoutError()
        if msg is INVALID:
            raise Error("Invalid response from service")
        if msg.msg_type is MsgType.CLOSE:
            raise Error("Connection closed by service")
        return msg
    def close(self) -> None:
        """Close connection to service.
        """
        self.protocol.send_close(self.channel, self.session)
        self.channel.discard_session(self.session)
        self.session = None
        self.channel = None
    @property
    def connected(self) -> bool:
        """True if client is connected to service.
        """
        return (self.session is not None) and (self.session.greeting is not None)



# SPDX-FileCopyrightText: 2019-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/service.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Services
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.
# pylint: disable=R0903

"""Saturnin base module for implementation of Firebird Butler Services
"""

from __future__ import annotations
from typing import cast, Final
from abc import abstractmethod
import uuid
import zmq
from firebird.base.config import ListOption
from saturnin.base import ZMQAddress, RouterChannel, ComponentConfig, ServiceDescriptor
from saturnin.component.micro import MicroService
from saturnin.protocol.fbsp import FBSPService

#: Channel name
SVC_CHN: Final[str] = 'service'

class ServiceConfig(ComponentConfig):
    """Base data provider/consumer microservice configuration.

    Arguments:
        name: Default configuration section name (service name)
    """
    def __init__(self, name: str):
        super().__init__(name)
        #: Service endpoint addresses
        self.endpoints: ListOption = \
            ListOption('endpoints', ZMQAddress, "List of service endpoints", required=True)

class Service(MicroService):
    """Base Firebird Butler Service.
    """
    def __init__(self, zmq_context: zmq.Context, descriptor: ServiceDescriptor, *,
                 peer_uid: uuid.UUID=None):
        """
        Arguments:
            zmq_context: ZeroMQ Context.
            descriptor: Service descriptor.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
        """
        super().__init__(zmq_context, descriptor, peer_uid=peer_uid)
        #: Channel for communication with service clients.
        self.svc_channel: RouterChannel = None
    def initialize(self, config: ServiceConfig) -> None:
        """Verify configuration and assemble service structural parts.

        Arguments:
            config: Service configuration.
        """
        super().initialize(config)
        # Set endpoints this service binds
        self.endpoints[SVC_CHN] = config.endpoints.value.copy()
        #: Service protocol
        service = FBSPService(service=self.descriptor, peer=self.peer)
        service.log_context = self.logging_id
        self.svc_channel = self.mngr.create_channel(RouterChannel, SVC_CHN, service,
                                                    routing_id=self.peer.uid.hex.encode('ascii'),
                                                    sock_opts={'maxmsgsize': 52428800,
                                                               'rcvhwm': 500,
                                                               'sndhwm': 500,})
        self.register_api_handlers(service)
    @abstractmethod
    def register_api_handlers(self, service: FBSPService) -> None:
        """Called by `.initialize()` for registration of service API handlers and FBSP
        service event handlers.

        Arguments:
            service: Service instance
        """
    def start_activities(self) -> None:
        """Start normal service activities.

        Must raise an exception when start fails.
        """
        super().start_activities()
        self.svc_channel.set_wait_in(True)
    def stop_activities(self) -> None:
        """Stop service activities.

        Calls `.FBSPService.close` and disables receiving incoming messages on the channel.
        """
        super().stop_activities()
        cast(FBSPService, self.svc_channel.protocol).close(self.svc_channel)
        self.svc_channel.set_wait_in(False)



# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/daemons.py
# DESCRIPTION:    Saturnin daemon commands
# CREATED:        16.02.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=R0912, R0913, R0914, R0915

"""Saturnin daemon commands


"""

from __future__ import annotations
from typing import Dict, List
import datetime
import psutil
import typer
from rich.table import Table
from rich.text import Text
from rich import box
from saturnin.base import directory_scheme
from saturnin.component.recipe import RecipeInfo, recipe_registry
from saturnin.lib.console import console
from saturnin.lib import daemon

#: Typer command group for daemon management commands
app = typer.Typer(rich_markup_mode="rich", help="Saturnin daemons.")

def get_first_line(text: str) -> str:
    """Returns first non-empty line from argument.
    """
    return text.strip().split('\n')[0]

def get_running_daemons() -> Dict[int, str]:
    """Returns dictionary with running daemons: key=pid, value=recipe_name
    """
    result = {}
    for pid_file in directory_scheme.pids.glob('*.pid'):
        recipe_name = pid_file.read_text()
        pid = int(pid_file.stem)
        if psutil.pid_exists(pid):
            result[pid] = recipe_registry.get(recipe_name)
        else:
            pid_file.unlink()
    return result

def pid_completer(ctx, args, incomplete) -> List:
    """Click completer for PIDs of running Saturnin daemons.
    """
    return [(str(pid), recipe.name if recipe else "")
            for pid, recipe in get_running_daemons().items()]

def format_size(size: int, decimals: int=2, binary_system: bool=True) -> str:
    """Returns "humanized" size.
    """
    if binary_system:
        units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB']
        largest_unit = 'YiB'
        step = 1024
    else:
        units = ['B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB']
        largest_unit = 'YB'
        step = 1000

    for unit in units:
        if size < step:
            return ('%.' + str(decimals) + 'f %s') % (size, unit)
        size /= step

    return ('%.' + str(decimals) + 'f %s') % (size, largest_unit)

@app.command()
def list_daemons() -> None:
    """List running Saturnin daemons.
    """
    daemons = get_running_daemons()
    #
    if daemons:
        table = Table(title='Running daemons', box=box.ROUNDED)
        table.add_column('PID', style='cyan')
        table.add_column('Status', style='bold yellow')
        table.add_column('Recipe', style='green')
        table.add_column('Description', style='white')
        recipe: RecipeInfo = None
        for pid, recipe in daemons.items():
            if recipe is None:
                table.add_row(str(pid), psutil.Process(pid).status(), 'UNKNOWN',
                              "[warning]Recipe used by daemon was uninstalled")
            else:
                table.add_row(str(pid), psutil.Process(pid).status(), recipe.name,
                              get_first_line(recipe.description))
        console.print(table)
    else:
        console.print("There are no Saturnin demons running.")


@app.command()
def show_daemon(pid: int=typer.Argument(..., metavar='PID', autocompletion=pid_completer)) -> None:
    """Show information about running Saturnin daemon.
    """
    if not psutil.pid_exists(pid):
        console.print_error("Process not found")
        return
    proc = psutil.Process(pid)
    data = proc.as_dict(ad_value='[important]N/A[/]')
    created = datetime.datetime.fromtimestamp(data['create_time'])
    run_time = datetime.datetime.now() - created
    table = Table(title=f"  Process [important]{pid}", box=box.ROUNDED, show_header=False,
                  title_justify='left')
    table.add_column('', style='green')
    table.add_column('')
    #
    table.add_row('Status:', Text(data['status'], style='important'))
    table.add_row('Created:', Text(created.strftime("%Y-%m-%d %H:%M:%S")))
    table.add_row('Run time:', Text(str(run_time)), end_section=True)

    table.add_row('# threads:', Text(str(data['num_threads'])))
    if 'num_handles' in data:
        table.add_row('# handles:', Text(str(data['num_handles'])))
    if 'num_fds' in data:
        table.add_row('# files:', Text(str(data['num_fds'])))
    children = proc.children()
    if children:
        parts = []
        line_sep = Text('\n')
        for child in children:
            parts.append(Text.assemble((str(child.pid), 'number'), ':', child.name(), ' - ', (child.status(), 'important')))
        table.add_row('Children:', line_sep.join(parts))
    table.add_row('# INET con.:', Text(str(len(data['connections']))), end_section=True)

    table.add_row('Name:', Text(data['name'], style='cyan'))
    table.add_row('Executable:', Text(data['exe']))
    table.add_row('Cmd. line:', Text(', '.join(data['cmdline'])))
    table.add_row('CWD:', Text(data['cwd']))
    table.add_row('User:', Text(data['username']), end_section=True)

    has_iowait = hasattr(data['cpu_times'], 'iowait')
    table.add_row('CPU user:', Text(str(data['cpu_times'].user)))
    table.add_row('CPU system:', Text(str(data['cpu_times'].system)),
                  end_section=not has_iowait)
    if has_iowait:
        table.add_row('CPU I/O wait:', Text(str(data['cpu_times'].iowait)), end_section=True)

    table.add_row('RSS (bytes):', Text(format_size(data['memory_info'].rss)))
    table.add_row('VMS (bytes):', Text(format_size(data['memory_info'].vms)), end_section=True)

    table.add_row('Read count:', Text(str(data['io_counters'].read_count)))
    table.add_row('Write count:', Text(str(data['io_counters'].write_count)))
    table.add_row('Bytes read:', Text(str(data['io_counters'].read_bytes)))
    table.add_row('Bytes written: ', Text(str(data['io_counters'].write_bytes)),
                  end_section=True)

    console.print(table)

@app.command()
def stop_daemon(pid: int=typer.Argument(..., metavar='PID', autocompletion=pid_completer)) -> None:
    """Stop running Saturnin daemon.
    """
    if not psutil.pid_exists(pid):
        console.print_error("Process not found")
        return
    daemon.stop_daemon(pid)



# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/single.py
# DESCRIPTION:    Single service controller and executor
# CREATED:        10.2.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=R0903

"""Single service controller and executor.


"""

from __future__ import annotations
from typing import List, Union, Tuple, Any
import uuid
import weakref
import warnings
from pathlib import Path
from configparser import ConfigParser, ExtendedInterpolation, DEFAULTSECT
import zmq
from firebird.base.logging import LoggingIdMixin, get_logger
from firebird.base.trace import TracedMixin
from saturnin.base import (Error, ChannelManager, SECTION_LOCAL_ADDRESS,
                           SECTION_NET_ADDRESS, SECTION_NODE_ADDRESS, SECTION_PEER_UID,
                           SECTION_SERVICE_UID, SECTION_SERVICE)
from .controller import ThreadController, DirectController, ServiceExecConfig, Outcome
from .registry import service_registry

class SingleController(LoggingIdMixin, TracedMixin):
    """Service controller that manages service executed directly or in separate thread.
    """
    def __init__(self, *, parser: ConfigParser=None, manager: ChannelManager=None,
                 direct: bool=False):
        """
        Arguments:
            controller_class: Inner controller class.
            parser: ConfigParser instance to be used for service configuration.
            manager: ChannelManager to be used.
            direct: Use DirectController or ThreadController.
        """
        #: Use DirectController instead ThreadController
        self.direct: bool = direct
        self.log_context = None
        #: Channel manager
        self.mngr: ChannelManager = manager
        self._ext_mngr: bool = manager is not None
        if manager is None:
            self.mngr = ChannelManager(zmq.Context.instance())
            self.mngr.log_context = weakref.proxy(self)
        #: ConfigParser with service configuration
        self.config: ConfigParser = \
            ConfigParser(interpolation=ExtendedInterpolation()) if parser is None else parser
        #: Service controller
        self.controller: Union[ThreadController, DirectController] = None
        #: Registry with ServiceDescriptors for services that could be run
        #
        self.config[SECTION_LOCAL_ADDRESS] = {}
        self.config[SECTION_NODE_ADDRESS] = {}
        self.config[SECTION_NET_ADDRESS] = {}
        self.config[SECTION_SERVICE_UID] = {}
        self.config[SECTION_PEER_UID] = {}
        # Defaults
        self.config[DEFAULTSECT]['here'] = str(Path.cwd())
        # Assign Agent IDs for available services
        self.config[SECTION_SERVICE_UID].update((sd.name, sd.uid.hex) for sd
                                                in service_registry)
        #
    def configure(self, *, section: str=SECTION_SERVICE) -> None:
        """
        Arguments:
            section: Configuration section with service specification.
        """
        svc_cfg: ServiceExecConfig = ServiceExecConfig(section)
        svc_cfg.load_config(self.config)
        svc_cfg.validate()
        peer_uid = uuid.uuid1()
        # Assign Peer ID to service section (instance)
        self.config[SECTION_PEER_UID][section] = peer_uid.hex
        #
        #
        controller_class = DirectController if self.direct else ThreadController
        if svc_cfg.agent.value in service_registry:
            self.controller = controller_class(service_registry[svc_cfg.agent.value],
                                               name=svc_cfg.name, peer_uid=peer_uid,
                                               manager=self.mngr)
        else:
            raise Error(f"Unknonw agent in section '{section}'")
    def start(self, *, timeout: int=10000) -> None:
        """Starts service.

        Arguments:
            timeout: Timeout for starting the service. `None` (infinity), or a floating
                     point number specifying a timeout for the operation in seconds (or
                     fractions thereof) [Default: 10s].

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When service does not start in time.
        """
        try:
            self.controller.configure(self.config, self.controller.name)
            self.controller.log_context = self.log_context
            self.controller.start(timeout=timeout)
        except:
            self.stop()
            raise
    def stop(self, *, timeout: int=10000) -> None:
        """Stop runing service.

        Arguments:
            timeout: Timeout for stopping the service. `None` (infinity), or a floating
                     point number specifying a timeout for the operation in seconds (or
                     fractions thereof) [Default: 10s].

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When service does not stop in time.
        """
        if not self.direct:
            try:
                self.controller.stop(timeout=timeout)
            except Exception as exc: # pylint: disable=W0703
                get_logger(self).error("Error while stopping the service: {args[0]}", exc)
                if self.controller.is_running():
                    warnings.warn(f"Stopping service {self.controller.name} failed, "
                                  f"service thread terminated", RuntimeWarning)
                    self.controller.terminate()
    def join(self, timeout=None) -> None:
        """Wait until service stops.

        Arguments:
            timeout: Floating point number specifying a timeout for the operation in
                     seconds (or fractions thereof).
        """
        self.controller.join(timeout)

class SingleExecutor():
    """Single service executor context manager.
    """
    def __init__(self, log_context: Any, *, direct: bool = False):
        """
        Arguments:
            log_context: Log context for this executor.
            direct: Use `.DirectController` (True) or `.ThreadController`.
        """
        #: Use DirectController instead ThreadController
        self.direct: bool = direct
        self.log_context = log_context
        #: Channel manager
        self.mngr: ChannelManager = None
        #: Controller
        self.controller: SingleController = None
    def __enter__(self) -> SingleExecutor:
        return self
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if self.mngr is not None:
            self.mngr.shutdown(forced=True)
        zmq.Context.instance().term()
    def configure(self, cfg_files: List[str], *, section: str=SECTION_SERVICE) -> None:
        """Executor configuration.

        Arguments:
          cfg_files: List of configuration files.
          section:   Configuration section name with service specification.
        """
        self.mngr = ChannelManager(zmq.Context.instance())
        self.mngr.log_context = self.log_context
        self.controller: SingleController = SingleController(manager=self.mngr,
                                                             direct=self.direct)
        self.controller.log_context = self.log_context
        self.controller.config.read(cfg_files)
        self.controller.configure(section=section)
    def run(self) -> List[Tuple[str, Outcome, List[str]]]:
        """Runs the service in main or separate thread.

        Returns:
          Tuple with (service_name, outcome, details).

        outcome: `.Outcome` of service execution.
        details: List of strings with additional outcome information (typically error text)
        """
        self.controller.start()
        result = None
        if not self.direct:
            try:
                self.controller.join()
                raise KeyboardInterrupt() # This, or direct call to executor.stop()
            except KeyboardInterrupt: # SIGINT
                self.controller.stop()
            finally:
                result = (self.controller.controller.outcome, self.controller.controller.details)
        return result



# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/data/metadata.py
# DESCRIPTION:    Module for work with importlib metadata
# CREATED:        23.2.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=W0611

"""Module for work with importlib metadata


"""

from __future__ import annotations
from typing import Generator, Optional
from importlib.metadata import (entry_points, EntryPoint, Distribution, distributions,
                                distribution)

def iter_entry_points(group: str, name: str=None) -> Generator[EntryPoint, None, None]:
    """Replacement for pkg_resources.iter_entry_points.

    Arguments:
        group: Entrypoint group name
        name:  Etrypoint name.

    When `name` is specified, returns only EntryPoint with such name. When `name` is not
    specified, returns all entry points in group.
    """
    for item in entry_points().get(group, []):
        if name is None or item.name == name:
            yield item

def get_entry_point_distribution(entry_point: EntryPoint) -> Optional[Distribution]:
    """Returns distribution that registered specified entry point, or None if distribution
    is not found. This function searches through all distributions, not only those that
    registered Saturnin components.

    Arguments:
      entry_point: Entry point for which the distribution is to be found.
    """
    for dis in (d for d in distributions() if d.entry_points):
        for entry in dis.entry_points:
            if entry.name == entry_point.name:
                return dis
    return None
